using System;
using System.Collections.Generic;
using System.Text;
using System.CodeDom;
using System.CodeDom.Compiler;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.Collections;
using System.IO;
using Microsoft.CSharp;
using Microsoft.VisualBasic;
using System.Security;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design.Validation;
using EnterpriseLibraryExtensions.ConfigurationConstsWriter.Properties;

namespace EnterpriseLibraryExtensions.ConfigurationConstsWriter
{
    public static class ConfigurationConstsWriter
    {
		/// <summary>
		/// Writes the codefile from the options provided.
		/// </summary>
		public static bool Write(ConfigurationNode node, ConfigurationConstsWriterSettings codeWriterSettings, string configurationPath)
		{
            //retrieve a CodeDomProvider for the specified CodingLanguage
            CodeDomProvider codeGenerator = CreateCodeGenerator(codeWriterSettings.CodingLanguage);

			//build the compileunit (or code tree)
            CodeCompileUnit unit = BuildCompileUnit(codeGenerator, codeWriterSettings);

			//write the sourcecode to disk
            string filePath = (Path.IsPathRooted(codeWriterSettings.FilePath)) ? codeWriterSettings.FilePath : Path.Combine(configurationPath, codeWriterSettings.FilePath);
            
            EnsureDirectoryExists(filePath);
            try
            {
                using (StreamWriter sw = new StreamWriter(filePath))
                {
                    codeGenerator.GenerateCodeFromCompileUnit(unit, sw, new CodeGeneratorOptions());
                    sw.Flush();
                    sw.Close();
                }
            }
            catch (UnauthorizedAccessException)
            {
                ServiceHelper.GetUIService(node.Site).ActivateNode(node);
                ServiceHelper.GetUIService(node.Site).ShowError(string.Format(Resources.UnabletoWriteConstants, filePath), Resources.UnabletoWriteConstantsTitle);
                return false;
            }
            catch (IOException)
            {
                ServiceHelper.GetUIService(node.Site).ActivateNode(node);
                ServiceHelper.GetUIService(node.Site).ShowError(string.Format(Resources.UnabletoWriteConstants, filePath), Resources.UnabletoWriteConstantsTitle);
                return false;
            }

            return true;
		}

        private static void EnsureDirectoryExists(string filepath)
        {
            string directory = Path.GetDirectoryName(filepath);
            if (!Directory.Exists(directory))
            {
                EnsureDirectoryExists(directory);
                Directory.CreateDirectory(directory);
            }
        }

		/// <summary>
		/// Builds the CodeDom for the constants file
		/// </summary>
		/// <returns></returns>
        private static CodeCompileUnit BuildCompileUnit(CodeDomProvider codeGen, ConfigurationConstsWriterSettings codeWriterSettings)
		{
			CodeCompileUnit ret = new CodeCompileUnit();
            CodeNamespace namespaceDecl = new CodeNamespace(codeWriterSettings.Namespace);
            CodeTypeDeclaration typeDecl = new CodeTypeDeclaration(codeWriterSettings.Class);
			
			CodeConstructor ctor = new CodeConstructor();
			ctor.Attributes = MemberAttributes.Private;
			typeDecl.Members.Add(ctor);
			
			Hashtable addedTypesByPath = new Hashtable();
            foreach (string nodePath in codeWriterSettings.ConstantNameByNodePath.AllKeys)
			{
                string nodeName = nodePath.Substring(nodePath.LastIndexOf('/') + 1);
                string constPath = codeWriterSettings.ConstantNameByNodePath[nodePath].Value;
                AddConstant(codeGen, typeDecl, constPath, nodeName, "", addedTypesByPath);
			}

			namespaceDecl.Types.Add(typeDecl);
			ret.Namespaces.Add(namespaceDecl);

			return ret;
		}

		private static void AddConstant(CodeDomProvider codeGen, CodeTypeDeclaration parentClass, string path, string constValue, string currentPath, Hashtable addedTypesByPath)
		{
            int pathSepperatorIndex = path.IndexOfAny(new char[] { '.', '/', '\\' });
			if(pathSepperatorIndex != -1)
			{
				string pathPart = path.Substring(0, pathSepperatorIndex);
				string pathRemainder = path.Substring(pathSepperatorIndex +1);
				string pathPartAsIdentifier = CreateIdentifier(pathPart, true);
				string nestedTypePath = string.Concat(currentPath, ".", pathPartAsIdentifier);

				CodeTypeDeclaration nestedClass = addedTypesByPath[nestedTypePath] as CodeTypeDeclaration;

				
				if (nestedClass == null)
				{
					nestedClass = new CodeTypeDeclaration(pathPartAsIdentifier);
					nestedClass.Attributes = MemberAttributes.Public;

					CodeConstructor ctor = new CodeConstructor();
					ctor.Attributes = MemberAttributes.Private;
					nestedClass.Members.Add(ctor);

					parentClass.Members.Add(nestedClass);

					addedTypesByPath.Add(nestedTypePath, nestedClass);

				}

				//recurse!
				AddConstant(codeGen, nestedClass, pathRemainder, constValue, nestedTypePath, addedTypesByPath);

			} 
			else 
			{
				string constName = CreateIdentifier(path, true);
				
				CodeMemberField constDecl = new CodeMemberField(typeof(string), constName);
				constDecl.Attributes = MemberAttributes.Const | MemberAttributes.Public;
				constDecl.InitExpression = new CodePrimitiveExpression(constValue);

				parentClass.Members.Add(constDecl);
			}
		}

		public static string CreateIdentifier(string inName, bool cammelCase)
		{
            StringBuilder tmpRet = new StringBuilder();
			bool NextIsCapital = cammelCase;
			foreach(char ThisChar in inName.ToCharArray())
			{
				if (((ThisChar > 47) && (ThisChar < 58)) || ((ThisChar > 64) && (ThisChar < 91)) || ((ThisChar > 96) && (ThisChar < 123)))
				{
					if (NextIsCapital == true)
					{
                        tmpRet.Append(Char.ToUpper(ThisChar));
						NextIsCapital = false;
					} 
					else 
					{
						tmpRet.Append(ThisChar);
					}
				} 
				else 
				{
					NextIsCapital = true;
				}
			}
			return tmpRet.ToString();
		}

		/// <summary>
		/// returns a ICodeGenerator based on the specified <c ref="CodingLanguage"/>.
		/// returns null when none is found
		/// </summary>
		private static CodeDomProvider CreateCodeGenerator(CodingLanguage lang)
		{
            return CodeDomProvider.CreateProvider(lang.ToString());
		}
    }
}
