using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design.Validation;
using System.Runtime.Serialization;
using System.Collections;
using System.Drawing.Design;

namespace EnterpriseLibraryExtensions.EnvironmentalOverrides
{
    [TypeConverter(typeof(EnvironmentalOverride.EnvironmentalOverrideTypeConverter))]
    [Serializable]
    public class EnvironmentalOverride : ICustomTypeDescriptor, ISerializable
    {
        Hashtable _backupValue = new Hashtable();
        Hashtable _copiedProperties = new Hashtable();
        PropertyDescriptorCollection _properties;
        ConfigurationNode _parentNode;
        bool _useParentProperties = true;

        public EnvironmentalOverride()
        {
        }

        public EnvironmentalOverride(SerializationInfo info, StreamingContext context)
		{
			_useParentProperties = info.GetBoolean("_useParentProperties");
            _copiedProperties = (Hashtable)info.GetValue("_copiedProperties", typeof(Hashtable));
		}

        public EnvironmentalOverride(ConfigurationNode parentNode, bool useParentProperties)
        {
            _parentNode = parentNode;
            _useParentProperties = useParentProperties;
        }

        public object GetBackupValue(string key)
        {
            if (_backupValue.ContainsKey(key))
            {
                return _backupValue[key];
            }
            else
            {
                return PropertyDescriptorWrapper.NoValuePlaceholder;
            }
        }

        public void SetBackupValue(string key, object backup)
        {
            _backupValue[key] = backup;
        }

        public void SetObjectToCopy(ConfigurationNode parentNode)
        {
            _parentNode = parentNode;
        }

        public bool UseParentProperties
        {
            get { return _useParentProperties; }
        }

        public ConfigurationNode ParentNode
        {
            get { return _parentNode; }
        }

        public static EnvironmentalOverride Default
        {
            get { return new EnvironmentalOverride(null, true); }
        }

        private object GetOverriddenValue(string propertyName)
        {
            return _copiedProperties[propertyName];
        }

        private void SetOverriddenValue(string propertyName, object value)
        {
            _copiedProperties[propertyName] = value;
        }

        public void SetOverridenValueWithBackup()
        {
            foreach (PropertyDescriptorWrapper wrappedProperty in GetProperties())
            {
                wrappedProperty.SetNewValueAndBackupOld(_parentNode);
            }
        }


        internal void SetBackupBack()
        {
            foreach (PropertyDescriptorWrapper wrappedProperty in GetProperties())
            {
                wrappedProperty.SetBackupBack(_parentNode);
            }
        }

        #region ICustomTypeDescriptor Members

        public TypeConverter GetConverter()
        {
            return TypeDescriptor.GetConverter(_parentNode, true);
        }

        public EventDescriptorCollection GetEvents(Attribute[] attributes)
        {
            return TypeDescriptor.GetEvents(_parentNode, attributes, true);
        }

        EventDescriptorCollection System.ComponentModel.ICustomTypeDescriptor.GetEvents()
        {
            return TypeDescriptor.GetEvents(_parentNode, true);
        }

        public string GetComponentName()
        {
            return TypeDescriptor.GetComponentName(_parentNode, true);
        }

        public object GetPropertyOwner(PropertyDescriptor pd)
        {
            return _parentNode;
        }

        public AttributeCollection GetAttributes()
        {
            return TypeDescriptor.GetAttributes(_parentNode, true);
        }

        public PropertyDescriptorCollection GetProperties(Attribute[] attributes)
        {
            //if (_properties == null)
            //{
                PropertyDescriptorCollection collectedProperties = TypeDescriptor.GetProperties(_parentNode, true);
                PropertyDescriptorCollection propertiesToOverride = SanitizePropertiesToCopy(_parentNode, collectedProperties);
                _properties = WrapProperties(propertiesToOverride);
            //}
            return _properties;
        }

        public PropertyDescriptorCollection GetProperties()
        {
            return GetProperties(new Attribute[0]);
        }

        private PropertyDescriptorCollection WrapProperties(PropertyDescriptorCollection propertiesToOverride)
        {
            List<PropertyDescriptor> readOnlyProperties = new List<PropertyDescriptor>();
            foreach (PropertyDescriptor prop in propertiesToOverride)
            {
                PropertyDescriptorWrapper pdw = new PropertyDescriptorWrapper(this, prop.DisplayName, prop);
                if (_copiedProperties != null && _copiedProperties.ContainsKey(pdw.Key))
                {
                    pdw.SetValue(_parentNode, _copiedProperties[pdw.Key]);
                }
                else if (!_useParentProperties)
                {
                    pdw.SetValue(_parentNode, prop.GetValue(_parentNode));
                }
                readOnlyProperties.Add(pdw);
            }
            propertiesToOverride = new PropertyDescriptorCollection(readOnlyProperties.ToArray());
            return propertiesToOverride;
        }

        private static PropertyDescriptorCollection SanitizePropertiesToCopy(object o, PropertyDescriptorCollection collectedProperties)
        {
            List<PropertyDescriptor> propertiesToReturn = new List<PropertyDescriptor>();

            foreach (PropertyDescriptor prop in collectedProperties)
            {
                if (prop.PropertyType != typeof(EnvironmentalOverride)
                    && prop.Name != "Name"  
                    && prop.IsBrowsable
                    && !prop.Attributes.Contains(new EnvironmentalOverridePropertyAttribute(false)))
                {
                    object propertyValue = prop.GetValue(o);
                    TypeConverter converter = prop.Converter;
                    
                    if (converter == null) converter = (propertyValue != null) ? TypeDescriptor.GetConverter(propertyValue) : null;
                    object editor  = prop.GetEditor(typeof(UITypeEditor));

                    if (editor == null || editor.GetType() != typeof(ReferenceEditor))
                    {
                        if (converter == null || !converter.GetPropertiesSupported())
                            propertiesToReturn.Add(prop);
                    }
                }
            }

            return new PropertyDescriptorCollection(propertiesToReturn.ToArray());
        }

        public object GetEditor(Type editorBaseType)
        {
            return TypeDescriptor.GetEditor(_parentNode, editorBaseType, true);
        }

        public PropertyDescriptor GetDefaultProperty()
        {
            return TypeDescriptor.GetDefaultProperty(_parentNode, true);
        }

        public EventDescriptor GetDefaultEvent()
        {
            return TypeDescriptor.GetDefaultEvent(_parentNode, true);
        }

        public string GetClassName()
        {
            return TypeDescriptor.GetClassName(_parentNode, true);
        }

        #endregion

        #region EnvironmentalOverrideTypeConverter
        public class EnvironmentalOverrideTypeConverter : ExpandableObjectConverter
        {
            private const string OverridePropertiesCaption = "Override properties";
            private const string ParentPropertiesCaption = "Dont override properties";

            public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
            {
                return true;
            }

            public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
            {
                return true;
            }

            public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
            {
                return true;
            }

            public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
            {
                return true;
            }

            public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
            {
                EnvironmentalOverride current = value as EnvironmentalOverride;
                if (current != null)
                {
                    if (!current.UseParentProperties)
                    {
                        return OverridePropertiesCaption;
                    }
                }
                return ParentPropertiesCaption;
            }

            public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
            {
                EnvironmentalOverride current = context.PropertyDescriptor.GetValue(context.Instance) as EnvironmentalOverride;
                if (current != null)
                {
                    if (value != null && string.Compare(value.ToString(), OverridePropertiesCaption) == 0)
                    {
                        return new EnvironmentalOverride(current.ParentNode, false);
                    }

                    return new EnvironmentalOverride(current.ParentNode, true);
                }
                return null;
            }


            public override System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
            {
                EnvironmentalOverride current = context.PropertyDescriptor.GetValue(context.Instance) as EnvironmentalOverride;

                EnvironmentalOverride[] standardValues = new EnvironmentalOverride[2];
                standardValues[0] = new EnvironmentalOverride(current.ParentNode, true);
                standardValues[1] = new EnvironmentalOverride(current.ParentNode, false);

                return new StandardValuesCollection(standardValues);
            }
        }

        #endregion

        #region PropertyDescriptorWrapper
        public class PropertyDescriptorWrapper : PropertyDescriptor
        {
            internal static readonly object NoValuePlaceholder = new object();

            List<Attribute> validationAttributes = new List<Attribute>();
            
            object newValue = NoValuePlaceholder;
            PropertyDescriptor _parentDescriptor;
            readonly bool _isReadonly;
            readonly EnvironmentalOverride _override;
            readonly string _overrideName;

            public PropertyDescriptorWrapper(EnvironmentalOverride @override, string overrideName, PropertyDescriptor parentDescriptor)
                : base(parentDescriptor)
            {
                _parentDescriptor = parentDescriptor;
                _isReadonly = @override.UseParentProperties;
                _override = @override;
                _overrideName = overrideName;
            }

            public void SetNewValueAndBackupOld(object component)
            {
                if (newValue != NoValuePlaceholder)
                {
                    _override.SetBackupValue(Key, _parentDescriptor.GetValue(component));
                    _parentDescriptor.SetValue(component, newValue);
                }
            }

            public override object GetEditor(Type editorBaseType)
            {
                if (editorBaseType == typeof(UITypeEditor))
                {
                    UITypeEditor editor = _parentDescriptor.GetEditor(editorBaseType) as UITypeEditor;
                    if (editor == null)  editor = base.GetEditor(editorBaseType) as UITypeEditor;
                    if (editor != null)
                    {
                        return new WrappedEditor(editor);
                    }
                }
                return base.GetEditor(editorBaseType);
            }

            public override AttributeCollection Attributes
            {
                get
                {
                    if (_isReadonly)
                    {
                        return new AttributeCollection(new Attribute[0]);
                    }
                    return base.Attributes;
                }
            }


            public void SetBackupBack(object component)
            {
                if (_override.GetBackupValue(Key) != NoValuePlaceholder)
                {
                    _parentDescriptor.SetValue(component, _override.GetBackupValue(Key));
                    _override.SetBackupValue(Key, NoValuePlaceholder);
                }
            }

            public override bool CanResetValue(object component)
            {
                return _parentDescriptor.CanResetValue(component);
            }

            public override object GetValue(object component)
            {
                if (newValue != NoValuePlaceholder)
                {
                    return _override.GetOverriddenValue(Key);
                }
                return _parentDescriptor.GetValue(component);
            }

            public override void ResetValue(object component)
            {
                newValue = NoValuePlaceholder;
                _parentDescriptor.ResetValue(component);
            }

            public override void SetValue(object component, object value)
            {
                _override.SetOverriddenValue(Key, value);
                newValue = value;
            }

            public override bool ShouldSerializeValue(object component)
            {
                return _parentDescriptor.ShouldSerializeValue(component);
            }

            public override Type ComponentType
            {
                get { return _parentDescriptor.ComponentType; }
            }

            public override bool IsReadOnly
            {
                get { return (_isReadonly) ? true : _parentDescriptor.IsReadOnly; }
            }

            public override Type PropertyType
            {
                get { return _parentDescriptor.PropertyType; }
            }

            public string Key
            {
                get
                {
                    if (_parentDescriptor is IOverriddenPropertyNameProvider)
                    {
                        return ((IOverriddenPropertyNameProvider)_parentDescriptor).GetOverriddenPropertyKey(base.Name);
                    }
                    return base.Name;
                }
            }

            public override string DisplayName
            {
                get
                {
                    if (_parentDescriptor is IOverriddenPropertyNameProvider)
                    {
                        return ((IOverriddenPropertyNameProvider)_parentDescriptor).GetOverriddenPropertyName(base.Name);
                    }
                    return base.Name;
                }
            }

            public override TypeConverter Converter
            {
                get
                {
                    return _parentDescriptor.Converter;
                }
            }

            public override string Category
            {
                get
                {
                    return _parentDescriptor.Category;
                }
            }

            public override string Description
            {
                get
                {
                    return _parentDescriptor.Description;
                }
            }
        }

        #endregion

        #region ISerializable Members

        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("_useParentProperties", _useParentProperties);
            info.AddValue("_copiedProperties", _copiedProperties);
        }

        #endregion
    }
}
