using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.IO;

namespace EnterpriseLibraryExtensions.EnvironmentalOverrides
{
    public class EnvironmentalOverridesManager : ConfigurationDesignManager
    {
        private bool recursiveSave = false;

        IExtenderStorageService extenderStorageService = null;
        public override void Register(IServiceProvider serviceProvider)
        {
            extenderStorageService = serviceProvider.GetService(typeof(IExtenderStorageService)) as IExtenderStorageService;

            EnvironmentalOverridesCommandRegistrar.Register(serviceProvider);
        }

        public override void Save(IServiceProvider serviceProvider)
        {
            if (recursiveSave) return;
            IUIService uiService = ServiceHelper.GetUIService(serviceProvider);
            base.Save(serviceProvider);
            try
            {
                uiService.BeginUpdate();

                foreach (EnvironmentNode envNode in ServiceHelper.GetCurrentHierarchy(serviceProvider).FindNodesByType(typeof(EnvironmentNode)))
                {
                    ConfigurationApplicationNode rootNode = (ConfigurationApplicationNode)ServiceHelper.GetCurrentHierarchy(serviceProvider).RootNode;
                    string baseConfigurationFilePath = rootNode.ConfigurationFile;
                    string baseConfigurationDir = Path.GetDirectoryName(baseConfigurationFilePath);
                    string configurationFileName = Path.GetFileName(baseConfigurationFilePath);

                    string environmentalConfigurationFilePath = Path.Combine(baseConfigurationDir, envNode.ConfigPath);
                    environmentalConfigurationFilePath = Path.Combine(environmentalConfigurationFilePath, configurationFileName);
                    EnsureDirectoryExists(environmentalConfigurationFilePath);

                    Dictionary<ConfigurationNode, EnvironmentalOverride> overrides = extenderStorageService.GetAllExtendedPropertyValues<EnvironmentalOverride>(envNode, EnvironmentNode.OverridesPropertyName);
                    try
                    {
                        recursiveSave = true;
                        foreach (ConfigurationNode extendee in overrides.Keys)
                        {
                            overrides[extendee].SetObjectToCopy(extendee);
                            overrides[extendee].SetOverridenValueWithBackup();
                        }
                        rootNode.ConfigurationFile = environmentalConfigurationFilePath;

                        new SaveConfigurationApplicationNodeCommand(rootNode.Site, false).Execute(rootNode);
                    }
                    finally
                    {
                        foreach (ConfigurationNode extendee in overrides.Keys)
                        {
                            overrides[extendee].SetBackupBack();
                        }
                        rootNode.ConfigurationFile = baseConfigurationFilePath;

                        recursiveSave = false;
                    }
                }
            }
            finally
            {
                uiService.EndUpdate();
            }
        }

        protected override ConfigurationSectionInfo GetConfigurationSectionInfo(IServiceProvider serviceProvider)
        {
            ConfigurationNode rootNode = ServiceHelper.GetCurrentRootNode(serviceProvider);
            EnvironmentalOverridesNode overridesNode = null;
            if (null != rootNode) overridesNode = rootNode.Hierarchy.FindNodeByType(rootNode, typeof(EnvironmentalOverridesNode)) as EnvironmentalOverridesNode;
            EnvironmentalOverridesSettings overridesSettings = null;
            if (overridesNode != null)
            {
                overridesSettings = EnvironmentalOverridesSettingsBuilder.Build(serviceProvider, overridesNode);
            }
            return new ConfigurationSectionInfo(overridesNode, overridesSettings, EnvironmentalOverridesSettings.SectionName);
        
        }

        protected override void OpenCore(IServiceProvider serviceProvider, ConfigurationApplicationNode rootNode, System.Configuration.ConfigurationSection section)
        {
            if (section != null)
            {
                rootNode.AddNode(EnvironmentalOverridesNodeBuilder.Build(serviceProvider, section as EnvironmentalOverridesSettings));
            }
        }

        private static void EnsureDirectoryExists(string filepath)
        {
            string directory = Path.GetDirectoryName(filepath);
            if (!Directory.Exists(directory))
            {
                EnsureDirectoryExists(directory);
                Directory.CreateDirectory(directory);
            }
        }
    }
}
