using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.ComponentModel;
using System.Reflection;
using System.Configuration;
using System.Collections;
using EnterpriseLibraryExtensions.EnvironmentalOverrides;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design.Validation;
using System.Diagnostics;

namespace EnterpriseLibraryExtensions.NakedNode
{


    public abstract class CollectionItem<TConfigurationElement> : CollectionItem
        where TConfigurationElement : ConfigurationElement
    {
        private Hashtable _propertyValues = new Hashtable();
        private static List<string> _knownProperties = new List<string>();
        private static List<string> _keyFields = new List<string>();
        private static List<string> _requiredFields = new List<string>();
        private static readonly List<ConfigurationPropertyAttribute> _configurationProperties = null;
        private static Dictionary<string, List<ConfigurationValidatorBase>> _propertiesWithValidation = new Dictionary<string, List<ConfigurationValidatorBase>>();
        private static Dictionary<string, string> _configurationPropertyNameMap;

        static CollectionItem()
        {
            _configurationPropertyNameMap = Utility.GetConfigurationPropertyNameMap(typeof(TConfigurationElement));
            _configurationProperties = Utility.GetConfigurationProperties(typeof(TConfigurationElement));
            _configurationProperties.ForEach(delegate(ConfigurationPropertyAttribute property)
            {
                if (property.IsKey) _keyFields.Add(property.Name);
                if (property.IsRequired) _requiredFields.Add(property.Name);
                _knownProperties.Add(property.Name);
            });

            foreach (PropertyInfo property in typeof(TConfigurationElement).GetProperties())
            {
                if (_configurationPropertyNameMap.ContainsKey(property.Name))
                {
                    string propertyName = _configurationPropertyNameMap[property.Name];
                    _propertiesWithValidation[propertyName] = new List<ConfigurationValidatorBase>();

                    foreach (ConfigurationValidatorAttribute validator in property.GetCustomAttributes(typeof(ConfigurationValidatorAttribute), true))
                    {
                        _propertiesWithValidation[propertyName].Add(validator.ValidatorInstance);
                    }
                }
            }
        }

        protected CollectionItem()
        {
        }

        protected CollectionItem(TConfigurationElement element)
        {
            foreach (PropertyInformation property in element.ElementInformation.Properties)
            {
                this[property.Name] = property.Value;
            }
        }

        public abstract TConfigurationElement CreateElement();

        public override Type GetConfigurationType()
        {
            return typeof(TConfigurationElement);
        }

        public override string GetCategoryCaption()
        {
            string nameString = string.Empty;

            switch (_keyFields.Count)
            {
                case 0:
                    break;

                case 1:
                    nameString = (this[_keyFields[0]] == null) ? string.Empty : this[_keyFields[0]].ToString();
                    Debug.WriteLine(string.Format("ugly categorykey = {0} for type {1}", nameString, GetType().Name));
                    break;

                default:
                    nameString = KeyString;
                    break;
            }

            return nameString;
        }

        public override object this[string propertyName]
        {
            get 
            {
                if (!_knownProperties.Contains(propertyName))
                    return UnkownPropertyGet(propertyName);
                
                return _propertyValues[propertyName];
            }
            set
            {
                if (!_knownProperties.Contains(propertyName))
                    UnkownPropertySet(propertyName, value);

                _propertyValues[propertyName] = value;
            }
        }

        protected virtual void UnkownPropertySet(string name, object value)
        {
            throw new InvalidOperationException(string.Format("Unkown property: {0}", name));
        }

        protected virtual object UnkownPropertyGet(string name)
        {
            throw new InvalidOperationException(string.Format("Unkown property: {0}", name));
        }

        public override void Validate(IList<ValidationError> errors)
        {
            foreach (string requiredPropertyName in _requiredFields)
            {
                if (_propertyValues[requiredPropertyName] == null || ((_propertyValues[requiredPropertyName] is string) && ((string)_propertyValues[requiredPropertyName] == string.Empty)))
                {
                    errors.Add(new ValidationError(Owner, "Collection", string.Format("Property {0} on item {1} cannot be empty.", requiredPropertyName, GetCategoryCaption())));
                }
            }
            if (errors.Count == 0)
            {
                foreach (string propertyName in _propertiesWithValidation.Keys)
                {
                    foreach (ConfigurationValidatorBase validator in _propertiesWithValidation[propertyName])
                    {
                        try
                        {
                            if (this[propertyName] != null)
                            {
                                validator.Validate(this[propertyName]);
                            }
                        }
                        catch (ConfigurationErrorsException configurationErrors)
                        {
                            errors.Add(new ValidationError(Owner, "Collection", configurationErrors.Message));
                        }
                        catch (Exception e)
                        {
                            errors.Add(new ValidationError(Owner, "Collection", e.Message));
                        }
                    }
                }
            }
            base.Validate(errors);
        }

        internal override string KeyString
        {
           get 
           {
                string keyString = string.Empty;
                _keyFields.ForEach(delegate(string keyFieldName) { keyString += string.Format("{0} = '{1}'", keyFieldName, this[keyFieldName]); });
                return keyString;
            }
        }
    }


    [TypeDescriptionProvider(typeof(NakedConfigurationNodeDescriptorProvider))]
    public abstract class CollectionItem //: IOverriddenPropertyNameProvider
    {
        private int _index;
        private ConfigurationNode _owner;

        [Browsable(false)]
        public int Index
        {
            get { return _index; }
            set { _index = value; }
        }

        public abstract string GetCategoryCaption();

        public abstract Type GetConfigurationType();

        //public virtual string GetOverriddenPropertyName(string propertyName)
        //{
        //    return string.Format("{0}. {1} {2}", Index, GetCategoryCaption(), propertyName);
        //}

        //public virtual string GetOverriddenPropertyKey(string propertyName)
        //{
        //    return string.Format("{0} {1}", GetCategoryCaption(), propertyName);
        //}

        public abstract object this[string propertyName]
        {
            get;
            set;
        }

        public virtual void Validate(IList<ValidationError> errors)
        {
        }

        public ConfigurationNode Owner
        {
            get { return _owner; }
            set { _owner = value; }
        }

        internal abstract string KeyString
        {
            get;
        }

    }

    public class NakedConfigurationNodeCollection<TConfigurationData, TCollectionItem> : NakedConfigurationNode<TConfigurationData>, INakedCollectionNode
        where TConfigurationData : new() 
        where TCollectionItem : CollectionItem, new()
    {
        List<TCollectionItem> _collection = new List<TCollectionItem>();
        protected bool EmitClear = false;
        private bool CanClear = true;

        public NakedConfigurationNodeCollection(TConfigurationData data, bool emitClear, bool canClear)
            :this(data, emitClear)
        {
            CanClear = canClear;
        }

        public NakedConfigurationNodeCollection(TConfigurationData data, bool emitClear)
            :base(data)
        {
            EmitClear = emitClear;
        }

        public NakedConfigurationNodeCollection()
            : base(new TConfigurationData())
        {
        }

        public virtual void CopyTo(object target)
        {
        }

        protected override void AddUnionedObjects()
        {
            if (CanClear)
            {
                AddUnionedObject("_inherit", new InheritCollectionItemsProperty(this));
            }
        }

        #region INakedCollectionNode Members

        public void AddCollectionItem(TCollectionItem collectionItem)
        {
            collectionItem.Index = _collection.Count;
            collectionItem.Owner = this;
            _collection.Add(collectionItem);
        }

        public virtual void AddCollectionItem()
        {
            TCollectionItem newItem = new TCollectionItem();
            AddCollectionItem(newItem);
        }

        public void RemoveCollectionItem(object collectionItem)
        {
            _collection.Remove((TCollectionItem)collectionItem);
            foreach (TCollectionItem item in _collection)
            {
                item.Index = _collection.IndexOf(item);
            }
        }

        public override void Validate(IList<ValidationError> errors)
        {
            List<String> foundKeyStrings = new List<string>();
            foreach (CollectionItem collectionItem in Collection)
            {
                if (foundKeyStrings.Contains(collectionItem.KeyString))
                {
                    errors.Add(new ValidationError(this, "Collection", string.Format("Duplicate key ({0}).", collectionItem.KeyString)));
                    
                }
                foundKeyStrings.Add(collectionItem.KeyString);
                collectionItem.Validate(errors);
            }
        }

        public IList Collection
        {
            get { return _collection; }
        }

        public string GetCollectionItemHeader(IList collection, object collectionItem)
        {
            return string.Format("{0}. {1}", ((TCollectionItem)collectionItem).Index, ((TCollectionItem)collectionItem).GetCategoryCaption());
        }

        #endregion

        protected override void InitializeProperties(IPropertyAttributeService propertyAttributeService)
        {
            TCollectionItem collectionItem = new TCollectionItem();
            base.InitializeProperties(propertyAttributeService);
            propertyAttributeService.SetDefaultAttributes(collectionItem.GetConfigurationType());

        }

        private class InheritCollectionItemsProperty
        {
            NakedConfigurationNodeCollection<TConfigurationData, TCollectionItem> _collection;

            public InheritCollectionItemsProperty(NakedConfigurationNodeCollection<TConfigurationData, TCollectionItem> collection)
            {
                _collection = collection;
            }

            [Category("(behaviour)")]
            public bool InheritCollectionItems
            {
                get { return !_collection.EmitClear; }
                set { _collection.EmitClear = !value; }
            }
        }
    }

    public class NakedConfigurationNode<TConfigurationData> : NakedConfigurationNode 
        where TConfigurationData : new()
    {
        public NakedConfigurationNode(TConfigurationData data)
            :base(data)
        {
        }

        public NakedConfigurationNode()
            : base(new TConfigurationData())
        {
        }

        public new TConfigurationData TheObject
        {
            get { return (TConfigurationData)base.TheObject; }
        }
    }

    [TypeDescriptionProvider(typeof(NakedConfigurationNodeDescriptorProvider))]
    public class NakedConfigurationNode: ConfigurationNode, ISitedConfigurationNodeSink
    {
        object _data;
        Dictionary<string, object> _unionedObjects = new Dictionary<string, object>();

        public NakedConfigurationNode(object data)
        {
            _data = data;
        }

        protected void AddUnionedObject(string name, object unionedObject)
        {
            if (_unionedObjects.ContainsKey(name))
            {
                return;
            }
            _unionedObjects.Add(name, unionedObject);
        }

        public object GetUnionedObjectByName(string name)
        {
            if (_unionedObjects.ContainsKey(name))
            {
                return _unionedObjects[name];
            }
            return null;
        }

        public object TheObject
        {
            get { return _data; }
        }

        public object[] UnionedObjects
        {
            get 
            { 
                object[] objects = new object[_unionedObjects.Count];
                _unionedObjects.Values.CopyTo(objects, 0);
                return objects;
            }
        }

        protected virtual void InitializeProperties(IPropertyAttributeService propertyAttributeService)
        {
            Type objectType = TheObject.GetType();
            propertyAttributeService.SetDefaultAttributes(objectType);
        }

        protected virtual void InitializeCommands(NakedNodeCommandRegistrar commandRegistrar)
        {
            Type objectType = TheObject.GetType();
            commandRegistrar.RegisterDefaultChildCommands(objectType, TheObject);
        }

        protected virtual void AddUnionedCollections()
        {
        }

        protected virtual void AddUnionedObjects()
        {
        }

        [ReadOnly(true)]
        public override string Name
        {
            get{return base.Name;}
            set{base.Name = value;}
        }

        #region ISitedConfigurationNodeSink Members

        public virtual void Sited()
        {
            AddUnionedCollections();
            AddUnionedObjects();
            foreach (Object o in UnionedObjects)
            {
                INakedNodeFactory nodeFactory = (INakedNodeFactory) Site.GetService(typeof(INakedNodeFactory));
                nodeFactory.SetNodeTypeForDataType(o.GetType(), this.GetType());
            }

            InitializeProperties((IPropertyAttributeService)Site.GetService(typeof(IPropertyAttributeService)));
            InitializeCommands(new NakedNodeCommandRegistrar(Site));
            
        }

        public void UnSited()
        {
        }

        #endregion
    }
}
