using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Collections;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.Configuration;
using System.ComponentModel.Design;
using System.Reflection;
using EnterpriseLibraryExtensions.EnvironmentalOverrides;
using System.Drawing.Design;

namespace EnterpriseLibraryExtensions.NakedNode
{
    class NakedConfigurationNodeDescriptorProvider : TypeDescriptionProvider
    {
        public override ICustomTypeDescriptor GetTypeDescriptor(Type objectType, object instance)
        {
            NakedConfigurationNode nakedNode = instance as NakedConfigurationNode;
            if (nakedNode != null)
            {
                return new NakedConfigurationNodeTypeDescriptor(nakedNode, nakedNode.TheObject, nakedNode.UnionedObjects);
            }

            CollectionItem collectionItemNode = instance as CollectionItem;
            if (collectionItemNode != null)
            {
                return new NakedConfigurationNodeCollectionItemTypeDescriptor(collectionItemNode, collectionItemNode.GetConfigurationType());
            }

            return base.GetTypeDescriptor(objectType, instance);
        }

        class NakedConfigurationNodeCollectionItemTypeDescriptor : CustomTypeDescriptor
        {
            IPropertyAttributeService _propertyAttributeService;
            CollectionItem _collectionItem;
            Type _configurationElementType;

            public NakedConfigurationNodeCollectionItemTypeDescriptor(CollectionItem collectionItem, Type configurationElementType)
            {
                _propertyAttributeService = (IPropertyAttributeService)collectionItem.Owner.Site.GetService(typeof(IPropertyAttributeService));
                _configurationElementType = configurationElementType;
                _collectionItem = collectionItem;
            }

            public override PropertyDescriptorCollection GetProperties()
            {
                return GetProperties(new Attribute[0]);
            }

            public override PropertyDescriptorCollection GetProperties(Attribute[] attributes)
            {
                List<PropertyDescriptor> pds = new List<PropertyDescriptor>();
                foreach (PropertyInfo property in _configurationElementType.GetProperties())
                {
                    if (property.IsDefined(typeof(ConfigurationPropertyAttribute), true))
                    {
                        ConfigurationPropertyAttribute configurationProperty = property.GetCustomAttributes(typeof(ConfigurationPropertyAttribute), true)[0] as ConfigurationPropertyAttribute;
                        if (configurationProperty != null)
                        {
                            List<Attribute> attributesForProperty = _propertyAttributeService.GetAttributesForProperty(new Attribute[0], _configurationElementType, property.Name);
                            pds.Add(new NakedConfigurationNodeCollectionItemPropertyDescriptor(_collectionItem, property, configurationProperty, attributesForProperty));
                        }
                    }
                }

                return new PropertyDescriptorCollection(pds.ToArray());
            }
        }

        class NakedConfigurationNodeCollectionItemPropertyDescriptor : NakedConfigurationNodePropertyDescriptorBase
        {
            CollectionItem _collectionItem;
            ConfigurationPropertyAttribute _configurationProperty;

            public NakedConfigurationNodeCollectionItemPropertyDescriptor(CollectionItem collectionItem, PropertyInfo property, ConfigurationPropertyAttribute configurationProperty, List<Attribute> attributes)
                :base(property.Name, attributes, collectionItem.Owner.Site, property.PropertyType)
            {
                _configurationProperty = configurationProperty;
                _collectionItem = collectionItem;
            }

            public override Type ComponentType
            {
                get { return _collectionItem.GetType(); }
            }

            protected override object ActualGetValue(object component)
            {
                if (_collectionItem[_configurationProperty.Name] == null)
                {
                    _collectionItem[_configurationProperty.Name] = ConvertValueForSet(_configurationProperty.DefaultValue);
                }
                return _collectionItem[_configurationProperty.Name];
            }

            public override bool IsReadOnly
            {
                get
                {
                    return ((ReadOnlyAttribute)this.Attributes[typeof(ReadOnlyAttribute)]).IsReadOnly;
                }
            }

            public override void ResetValue(object component)
            {
            }

            protected override void ActualSetValue(object component, object value)
            {
                _collectionItem[_configurationProperty.Name] = value;
            }

            public override bool ShouldSerializeValue(object component)
            {
                return false;
            }
        }

        class NakedConfigurationNodeTypeDescriptor : CustomTypeDescriptor
        {
            IPropertyAttributeService _propertyAttributeService;
            IExtenderListService _extenderService;
            ConfigurationNode _node;

            object _primaryInstance;
            object[] _unionedInstances;

            public NakedConfigurationNodeTypeDescriptor(ConfigurationNode node, object primaryInstance, object[] unionedInstances)
            {
                _extenderService = (IExtenderListService) node.Site.GetService(typeof(IExtenderListService));
                _propertyAttributeService = (IPropertyAttributeService)node.Site.GetService(typeof(IPropertyAttributeService));
                _node = node;
                _primaryInstance = primaryInstance;
                _unionedInstances = unionedInstances;
            }

            public override PropertyDescriptorCollection GetProperties()
            {
                return GetProperties(new Attribute[0]);
            }

            public override PropertyDescriptorCollection GetProperties(Attribute[] attributes)
            {
                List<NakedConfigurationNodePropertyDescriptorBase> properties = new List<NakedConfigurationNodePropertyDescriptorBase>();

                CollectProperties(_primaryInstance, attributes, properties);
                CollectExtendedProperties(_primaryInstance, attributes, properties, new Attribute[0]);

                foreach (Object unionedObject in _unionedInstances)
                {
                    CollectProperties(unionedObject, attributes, properties);
                }

                if (_node is INakedCollectionNode)
                {
                    INakedCollectionNode collectionNode = (INakedCollectionNode)_node;
                    
                    foreach (object collectionItem in collectionNode.Collection)
                    {
                        CategoryAttribute categoryAttribute = new CategoryAttribute(collectionNode.GetCollectionItemHeader(collectionNode.Collection, collectionItem));
                        CollectProperties(collectionItem, attributes, properties, new Attribute[] { categoryAttribute });
                    }
                }

                return new PropertyDescriptorCollection(properties.ToArray());
            }

            private void CollectProperties(object theObject, Attribute[] attributes, List<NakedConfigurationNodePropertyDescriptorBase> properties)
            {
                CollectProperties(theObject, attributes, properties, new Attribute[0]);
            }

            private void CollectProperties(object theObject, Attribute[] attributes, List<NakedConfigurationNodePropertyDescriptorBase> properties, Attribute[] extraAttributes)
            {
                foreach (PropertyDescriptor propertyOnObject in TypeDescriptor.GetProperties(theObject, attributes, true))
                {
                    ArrayList foundAttributes = new ArrayList(propertyOnObject.Attributes);

                    List<Attribute> attributesForProperty = _propertyAttributeService.GetAttributesForProperty((Attribute[])foundAttributes.ToArray(typeof(Attribute)), propertyOnObject.ComponentType, propertyOnObject.Name);

                    attributesForProperty.AddRange(extraAttributes);
                    properties.Add(new NakedConfigurationNodePropertyDescriptor(_node, theObject, propertyOnObject.Name, propertyOnObject, attributesForProperty));
                }
            }

            private void CollectExtendedProperties(object theObject, Attribute[] attributes, List<NakedConfigurationNodePropertyDescriptorBase> properties, Attribute[] extraAttributes)
            {
                Assembly systemAsm = Assembly.GetAssembly(typeof(PropertyDescriptor));
                foreach (Type t in systemAsm.GetTypes())
                {
                    if (t.Name == "ReflectTypeDescriptionProvider")
                    {
                        MethodInfo mi = t.GetMethod("GetExtendedProperties", BindingFlags.Instance | BindingFlags.NonPublic);
                        Object i = Activator.CreateInstance(t, true);
                        PropertyDescriptorCollection listOfProperties = (PropertyDescriptorCollection)mi.Invoke(i, new object[] { _node });

                        foreach (PropertyDescriptor pd in listOfProperties)
                        {
                            List<Attribute> attributesOnExtendedProperty = new List<Attribute>();
                            attributesOnExtendedProperty.Add(new CategoryAttribute(pd.Category));
                            properties.Add(new NakedConfigurationNodePropertyDescriptor(_node, _node, pd.DisplayName, pd, attributesOnExtendedProperty));
                        }

                        return;
                    }
                }
            }
        }

        class NakedConfigurationNodePropertyDescriptor : NakedConfigurationNodePropertyDescriptorBase//, IOverriddenPropertyNameProvider
        {
            PropertyDescriptor _parentDescriptor;
            Object _theObject;
            ConfigurationNode _node;

            public NakedConfigurationNodePropertyDescriptor(ConfigurationNode node, object theObject, string propertyName, PropertyDescriptor parentDescriptor, List<Attribute> attributes)
                :base(propertyName, attributes, node.Site, parentDescriptor is NakedConfigurationNodePropertyDescriptorBase ? ((NakedConfigurationNodePropertyDescriptorBase)parentDescriptor).InternalPropertyType : parentDescriptor.PropertyType)
            {
                _node = node;
                _theObject = theObject;
                _parentDescriptor = parentDescriptor;
            }


            public override Type ComponentType
            {
                get { return _parentDescriptor.ComponentType; }
            }

            protected override object ActualGetValue(object component)
            {
                return _parentDescriptor.GetValue(_theObject);
            }

            public override bool IsReadOnly
            {
                get { return _parentDescriptor.IsReadOnly ; }
            }

            public override void ResetValue(object component)
            {
                _parentDescriptor.ResetValue(_theObject);
            }

            protected override void ActualSetValue(object component, object value)
            {
                _parentDescriptor.SetValue(_theObject, value);
            }

            public override bool ShouldSerializeValue(object component)
            {
                return false;
            }

            //#region IOverriddenPropertyNameProvider Members

            //public string GetOverriddenPropertyName(string propertyName)
            //{
            //    if (_theObject is IOverriddenPropertyNameProvider)
            //    {
            //        return ((IOverriddenPropertyNameProvider)_theObject).GetOverriddenPropertyName(propertyName);
            //    }
            //    return propertyName;
            //}

            //public string GetOverriddenPropertyKey(string propertyName)
            //{
            //    if (_theObject is IOverriddenPropertyNameProvider)
            //    {
            //        return ((IOverriddenPropertyNameProvider)_theObject).GetOverriddenPropertyKey(propertyName);
            //    }
            //    return propertyName;
            //}

            //#endregion
        }

        abstract class NakedConfigurationNodePropertyDescriptorBase : PropertyDescriptor
        {
            UITypeEditor _editor;
            TypeConverter _converter;
            TypeConverter _internalConverter;
            IUIService _uiService;
            Type _propertyType;
            Type _internalPropertyType;
            bool _refreshPropertyGridOnSet = false;
            
            protected NakedConfigurationNodePropertyDescriptorBase(string propertyName, List<Attribute> attributes, IServiceProvider serviceProvider, Type propertyType)
                :base(propertyName, attributes.ToArray())
            {
                RefreshPropertyGridAttribute refreshPropertyGrid = attributes.FindLast(delegate(Attribute a) { return a is RefreshPropertyGridAttribute; }) as RefreshPropertyGridAttribute;
                if (refreshPropertyGrid != null)
                {
                    _refreshPropertyGridOnSet = refreshPropertyGrid.RefreshPropertyGridOnSet;
                }

                _uiService = (IUIService)serviceProvider.GetService(typeof(IUIService));

                TypeAttribute typeAttribute = attributes.FindLast(delegate(Attribute a) { return a is TypeAttribute; }) as TypeAttribute;
                if (typeAttribute != null)
                {
                    _propertyType = typeAttribute.Type;
                    _internalConverter = Activator.CreateInstance(typeAttribute.TypeConverter) as TypeConverter;
                }
                else
                {
                    _propertyType = propertyType;
                }

                _internalPropertyType = propertyType;


                ITypeConverterService converterService = (ITypeConverterService)serviceProvider.GetService(typeof(ITypeConverterService));
                _converter = converterService.GetTypeConverterForType(propertyType);

                if (_converter == null)
                {
                    TypeConverterAttribute converterAttribute = attributes.FindLast(delegate(Attribute a) { return a is TypeConverterAttribute; }) as TypeConverterAttribute;
                    if (converterAttribute != null)
                    {
                        _converter = Activator.CreateInstance(Type.GetType(converterAttribute.ConverterTypeName)) as TypeConverter;
                    }
                }
                ITypeEditorService typeEditorService = (ITypeEditorService)serviceProvider.GetService(typeof(ITypeEditorService));
                _editor = typeEditorService.GetEditor(propertyType);
            }

            protected override void OnValueChanged(object component, EventArgs e)
            {
                base.OnValueChanged(component, e);

                if (_refreshPropertyGridOnSet)
                {
                    _uiService.RefreshPropertyGrid();
                }
            }

            public override TypeConverter Converter
            {
                get
                {
                    if (_converter != null) return _converter;

                    return base.Converter;
                }
            }


            public override void SetValue(object component, object value)
            {
                if (GetValue(component) != value)
                {
                    ActualSetValue(component, ConvertValueForSet(value));
                    OnValueChanged(component, EventArgs.Empty);
                }
            }

            protected object ConvertValueForSet(object value)
            {
                object newValue = value;
                if (value != null && !InternalPropertyType.IsAssignableFrom(value.GetType()))
                {
                    if (_internalConverter != null && _internalConverter.CanConvertFrom(value.GetType()))
                    {
                        newValue = _internalConverter.ConvertFrom(value);
                    }
                    else if (Converter.CanConvertFrom(value.GetType()))
                    {
                        newValue = Converter.ConvertFrom(value);
                    }
                    else
                    {
                        return null;
                    }
                }

                return newValue;
            }

            public override object GetValue(object component)
            {
                object value = ActualGetValue(component);
                if (value != null && !PropertyType.IsAssignableFrom(value.GetType()) && _internalConverter != null)
                {
                    value = _internalConverter.ConvertTo(value, PropertyType);
                }

                return value;
            }

            protected abstract void ActualSetValue(object component, object value);

            protected abstract object ActualGetValue(object component);

            public override Type PropertyType
            {
                get { return _propertyType; }
            }

            internal Type InternalPropertyType
            {
                get { return _internalPropertyType; }
            }

            public override object GetEditor(Type editorBaseType)
            {
                if (editorBaseType == typeof(UITypeEditor) && _editor != null)
                {
                    return _editor;
                }
                return base.GetEditor(editorBaseType);
            }

            public override bool CanResetValue(object component)
            {
                return true;
            }
        }
    }
}
