using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Configuration;
using System.ComponentModel;
using EnterpriseLibraryExtensions.EnvironmentalOverrides;
using System.Collections.ObjectModel;

namespace EnterpriseLibraryExtensions.NakedNode
{
    internal class PropertyAttributeService : IPropertyAttributeService
    {
        Dictionary<Type, List<Attribute>> attributesByType = new Dictionary<Type,List<Attribute>>();
        Dictionary<TypeAndPropertyName, List<Attribute>> attributesByProperty = new Dictionary<TypeAndPropertyName,List<Attribute>>();

        public PropertyAttributeService()
        {
            AddAttributeForAllProperties(typeof(ConfigurationElement), new BrowsableAttribute(false));
            AddAttributeForAllProperties(typeof(ConfigurationSection), new BrowsableAttribute(false));
            AddAttributeForAllProperties(typeof(ConfigurationSectionGroup), new BrowsableAttribute(false));
            AddAttributeForAllProperties(typeof(ConfigurationElementCollection), new BrowsableAttribute(false));
        }

        #region IPropertyAttributeService Members

        public void AddAttributeForAllProperties(Type type, Attribute att)
        {
            if (!attributesByType.ContainsKey(type))
            {
                attributesByType.Add(type, new List<Attribute>());
            }

            attributesByType[type].Add(att);
        }

        public void AddAttributeForProperty(Type type, string propertyName, Attribute att)
        {
            TypeAndPropertyName typeAndProperty = new TypeAndPropertyName(type, propertyName);
            if (!attributesByProperty.ContainsKey(typeAndProperty))
            {
                attributesByProperty.Add(typeAndProperty, new List<Attribute>());
            }

            attributesByProperty[typeAndProperty].Add(att);
        }

        public List<Attribute> GetAttributesForProperty(Attribute[] currentAttributes, Type type, string propertyName)
        {
            List<Attribute> attributes = new List<Attribute>(currentAttributes);
            if (attributesByType.ContainsKey(type))
            {
                attributes.AddRange(attributesByType[type].ToArray());
            }

            TypeAndPropertyName tp = new TypeAndPropertyName(type, propertyName);
            if (attributesByProperty.ContainsKey(tp))
            {
                attributes.AddRange(attributesByProperty[tp].ToArray());
            }

            return attributes;
        }

        public void SetDefaultAttributes(Type type)
        {
            AddAttributeForAllProperties(type, new CategoryAttribute("(general)"));
            foreach (PropertyInfo pi in type.GetProperties())
            {
                if (pi.PropertyType == typeof(ConfigurationElementCollectionType))
                {
                    AddAttributeForProperty(type, pi.Name, new BrowsableAttribute(false));
                }
                if (typeof(ConfigurationElement).IsAssignableFrom(pi.PropertyType))
                {
                    AddAttributeForProperty(type, pi.Name, new BrowsableAttribute(false));
                }

                if (pi.IsDefined(typeof(ConfigurationPropertyAttribute), true))
                {
                    ConfigurationPropertyAttribute configurationPropertyAttribute = pi.GetCustomAttributes(typeof(ConfigurationPropertyAttribute), true)[0] as ConfigurationPropertyAttribute;
                    if (configurationPropertyAttribute != null && configurationPropertyAttribute.IsKey)
                    {
                        AddAttributeForProperty(type, pi.Name, new RefreshPropertyGridAttribute(true));
                        AddAttributeForProperty(type, pi.Name, new EnvironmentalOverridePropertyAttribute(false));
                    }
                }
            }
        }

        #endregion

        #region TypeAndPropertyName
        private struct TypeAndPropertyName
        {
            Type _type;
            string _propertyName;

            public TypeAndPropertyName(Type type, string propertyName)
            {
                _type = type;
                _propertyName = propertyName;
            }
        }
        #endregion TypeAndPropertyName
    }

    public interface IPropertyAttributeService
    {
        void AddAttributeForAllProperties(Type type, Attribute att);
        void AddAttributeForProperty(Type type, string propertyName, Attribute att);

        List<Attribute> GetAttributesForProperty(Attribute[] currentAttributes, Type type, string propertyName);

        void SetDefaultAttributes(Type type);

    }
}
