using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.ComponentModel;
using System.Configuration;
using System.Drawing.Design;
using EnterpriseLibraryExtensions.Protection.Properties;
using System.Windows.Forms.Design;
using System.Windows.Forms;
using System.Collections;

namespace EnterpriseLibraryExtensions.Protection
{
    [ProvideProperty(ProtectProperty, typeof(ConfigurationNode))]
    [Image(typeof(ProtectionNode))]
    [SelectedImage(typeof(ProtectionNode))]
    [NodeInformation(typeof(ProtectionNode), "ProtectionNodeInfoHtml", "ProtectionNodeInfoTitle", typeof(Resources))]
    public class ProtectionNode : ConfigurationNode, IExtenderProvider, ISitedConfigurationNodeSink
    {
        static Dictionary<string, string> ConfigurationSectionNameByNodeTypeName = new Dictionary<string, string>();

        static ProtectionNode()
        {
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Caching.Configuration.Design.CacheManagerSettingsNode", "cachingConfiguration");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Configuration.Design.ConfigurationSourceSectionNode", "enterpriseLibrary.ConfigurationSource");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Configuration.Design.InstrumentationNode", "instrumentationConfiguration");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Security.Cryptography.Configuration.Design.CryptographySettingsNode", "securityCryptographyConfiguration");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Data.Configuration.Design.DatabaseSectionNode", "connectionStrings");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.ExceptionHandling.Configuration.Design.ExceptionHandlingSettingsNode", "exceptionHandling");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Logging.Configuration.Design.LoggingSettingsNode", "loggingConfiguration");
            ConfigurationSectionNameByNodeTypeName.Add("Microsoft.Practices.EnterpriseLibrary.Security.Configuration.Design.SecuritySettingsNode", "securityConfiguration");
            ConfigurationSectionNameByNodeTypeName.Add("EnterpriseLibraryExtensions.ConfigurationConstsWriter.ConfigurationConstsWriterNode", "enterpriseLibraryExtensions.ConfigurationConstsWriter");
            ConfigurationSectionNameByNodeTypeName.Add("EnterpriseLibraryExtensions.EnvironmentalOverrides.EnvironmentalOverridesNode", "enterpriseLibraryExtensions.EnvironmentalOverrides");
        }

        public const string ProtectProperty = "Protect";

        IExtenderStorageService _extenderStorageService;
        ICurrentSelectionService _currentSelectionService;

        private string _protectionProvider;

        public ProtectionNode()
            :this(new ProtectionSettings())
        {
        }

        public ProtectionNode(ProtectionSettings settings)
        {
            Name = settings.Name;
            _protectionProvider = string.IsNullOrEmpty(settings.ProtectionProvider) ? ProtectedConfiguration.DefaultProvider : settings.ProtectionProvider;
        }

        [TypeConverter(typeof(ProtectionProviderConverter))]
        [SRCategory("General", typeof(Resources))]
        [SRDescription("ProtectionProvider", typeof(Resources))]
        public string ProtectionProvider
        {
            get { return _protectionProvider; }
            set { _protectionProvider = value; }
        }

        #region IExtenderProvider Members

        [DefaultValue(false)]
        [SRDescription("Protect", typeof(Resources))]
        [SRCategory("Protection", typeof(Resources))]
        public bool GetProtect(ConfigurationNode configNode)
        {
            return _extenderStorageService.GetExtendedPropertyValue<string>(this, ProtectProperty, configNode) != null;
        }

        public void SetProtect(ConfigurationNode configNode, bool protect)
        {
            string configurationNodeTypeName = configNode.GetType().FullName;
            string sectionToProtect = null;
            if (ConfigurationSectionNameByNodeTypeName.ContainsKey(configurationNodeTypeName))
            {
                sectionToProtect = ConfigurationSectionNameByNodeTypeName[configurationNodeTypeName];
            }
            else if (configNode.GetType().IsDefined(typeof(ProtectableAttribute), false))
            {
                ProtectableAttribute protectable = configNode.GetType().GetCustomAttributes(typeof(ProtectableAttribute), false)[0] as ProtectableAttribute;
                sectionToProtect = protectable.SectionName;
            }
            if (sectionToProtect == null) return;

            _extenderStorageService.SetExtendedPropertyValue<string>(this, ProtectProperty, configNode, (protect) ? sectionToProtect : null);
        }

        public bool CanExtend(object extendee)
        {
            return (extendee is ConfigurationNode 
                && ((ConfigurationNode)extendee).Hierarchy == Hierarchy
                && _currentSelectionService.CurrentlySelectedNode == extendee
                && (    extendee.GetType().IsDefined(typeof(ProtectableAttribute), false)
                        || ConfigurationSectionNameByNodeTypeName.ContainsKey(extendee.GetType().FullName)));
        }

        #endregion

        #region ISitedConfigurationNodeSink Members

        public void Sited()
        {
            _extenderStorageService = (IExtenderStorageService) Site.GetService(typeof(IExtenderStorageService));
            _currentSelectionService = (ICurrentSelectionService)Site.GetService(typeof(ICurrentSelectionService));
        }

        public void UnSited()
        {
        }

        #endregion

        #region ProtectionProviderConverter
        class ProtectionProviderConverter : StringConverter
        {
            public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
            {
                return true;
            }

            public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
            {
                return true;
            }

            public override TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
            {
                List<string> stdValues = new List<string>();
                foreach (ProtectedConfigurationProvider p in ProtectedConfiguration.Providers)
                {
                    stdValues.Add(p.Name);
                }

                return new StandardValuesCollection(stdValues);
            }
        }
        #endregion
    }
}
