using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using EnterpriseLibraryExtensions.NakedNode;
using System.Web.Configuration;
using System.Configuration;
using System.Reflection;

namespace EnterpriseLibraryExtensions.SystemWeb
{
    public class SystemWebDesignManager : NakedNodeDesignManager
    {
        public SystemWebDesignManager()
        {
        }

        protected override void OpenCore(IServiceProvider serviceProvider, ConfigurationApplicationNode rootNode, ConfigurationSection section)
        {
            SystemWebNodeBuilder nodeBuilder = new SystemWebNodeBuilder(serviceProvider);
            nodeBuilder.BuildNode(rootNode);
        }

        public override void Save(IServiceProvider serviceProvider)
        {
            base.Save(serviceProvider);

            SystemWebSectionGroupNode systemWebNode = ServiceHelper.GetCurrentRootNode(serviceProvider).Hierarchy.FindNodeByType(typeof(SystemWebSectionGroupNode)) as SystemWebSectionGroupNode;
            if (systemWebNode != null)
            {
                new SystemWebSettingsBuilder(serviceProvider).SaveSettings(systemWebNode);
            }

        }

        public override void Register(IServiceProvider serviceProvider)
        {
            base.Register(serviceProvider);
            new SystemWebCommandRegistrar(serviceProvider).Register();
        }
    }

    public class SystemWebSettingsBuilder
    {
        IServiceProvider _serviceProvider;

        public SystemWebSettingsBuilder(IServiceProvider serviceProvider)
        {
            _serviceProvider = serviceProvider;
        }

        public void SaveSettings(SystemWebSectionGroupNode sysWebNode)
        {
            ExeConfigurationFileMap fileMap = new ExeConfigurationFileMap();
            fileMap.ExeConfigFilename = ServiceHelper.GetApplicationConfigurationFile(_serviceProvider);
            Configuration config = ConfigurationManager.OpenMappedExeConfiguration(fileMap, ConfigurationUserLevel.None);

            SystemWebSectionGroup systemWebSectionGroup = config.GetSectionGroup("system.web") as SystemWebSectionGroup;
            foreach (PropertyInfo sectionProperty in systemWebSectionGroup.GetType().GetProperties())
            {
                if (typeof(ConfigurationSection).IsAssignableFrom(sectionProperty.PropertyType)
                    && sectionProperty.GetIndexParameters().Length == 0
                    && sectionProperty.PropertyType != typeof(DefaultSection))
                {
                    ConfigurationSection webSection = sectionProperty.GetValue(systemWebSectionGroup, new object[0]) as ConfigurationSection;
                    if (webSection != null)
                    {
                        NakedConfigurationNode node = sysWebNode.Nodes[sectionProperty.Name] as NakedConfigurationNode;
                        if (node != null)
                        {
                            CopyProperties(node, node.TheObject as ConfigurationElement, webSection);
                        }
                        else
                        {
                            webSection.SectionInformation.RevertToParent();
                        }
                    }
                }
            }

            config.Save(ConfigurationSaveMode.Minimal);
        }

        private void CopyProperties(NakedConfigurationNode sourceNode, ConfigurationElement sourceElement, ConfigurationElement dest)
        {
            if (sourceElement == null || sourceNode == null || dest == null) return;

            Dictionary<string, PropertyInformation> DestPropertiesByName = new Dictionary<string, PropertyInformation>();
            foreach (PropertyInformation ci in dest.ElementInformation.Properties)
            {
                DestPropertiesByName.Add(ci.Name, ci);
            }

            foreach (PropertyInfo pi in sourceElement.GetType().GetProperties())
            {
                if (!pi.IsDefined(typeof(ConfigurationPropertyAttribute), true)) continue;

                ConfigurationPropertyAttribute configurationProperty = pi.GetCustomAttributes(typeof(ConfigurationPropertyAttribute), true)[0] as ConfigurationPropertyAttribute;
                PropertyInformation ci = sourceElement.ElementInformation.Properties[configurationProperty.Name];

                if (ci != null && DestPropertiesByName.ContainsKey(ci.Name))
                {
                    if (typeof(ConfigurationElement).IsAssignableFrom(DestPropertiesByName[ci.Name].Type))
                    {
                        NakedConfigurationNode childNode = sourceNode.Nodes[pi.Name] as NakedConfigurationNode;
                        if (childNode != null)
                        {
                            CopyProperties(childNode, childNode.TheObject as ConfigurationElement, DestPropertiesByName[ci.Name].Value as ConfigurationElement);
                            continue;
                        }

                        ConfigurationElement element = sourceNode.GetUnionedObjectByName(ci.Name) as ConfigurationElement;
                        if (element != null)
                        {
                            CopyProperties(sourceNode, element, DestPropertiesByName[ci.Name].Value as ConfigurationElement);
                        }

                    }
                    else
                    {
                        if (ci.Value != DestPropertiesByName[ci.Name].Value)
                        {
                            DestPropertiesByName[ci.Name].Value = ci.Value;
                        }
                    }
                }
            }

            if (sourceNode is INakedCollectionNode)
            {
                ((INakedCollectionNode)sourceNode).CopyTo(dest);
            }
        }
    }

    public class SystemWebNodeBuilder : NodeBuilder
    {
        INakedNodeFactory nakedNodeTypeFactory;
        IUICommandService commandService;

        public SystemWebNodeBuilder(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
            nakedNodeTypeFactory = (INakedNodeFactory) serviceProvider.GetService(typeof(INakedNodeFactory));
            commandService = (IUICommandService)serviceProvider.GetService(typeof(IUICommandService));
        }

        public void BuildNode(ConfigurationNode parentNode)
        {
            ExeConfigurationFileMap fileMap = new ExeConfigurationFileMap();
            fileMap.ExeConfigFilename = ServiceHelper.GetApplicationConfigurationFile(ServiceProvider);
            Configuration config = ConfigurationManager.OpenMappedExeConfiguration(fileMap, ConfigurationUserLevel.None);

            SystemWebSectionGroup systemWebSectionGroup = config.GetSectionGroup("system.web") as SystemWebSectionGroup;

            SystemWebSectionGroupNode systemWebNode = new SystemWebSectionGroupNode(systemWebSectionGroup);
            parentNode.AddNode(systemWebNode);

            foreach (PropertyInfo sectionProperty in systemWebSectionGroup.GetType().GetProperties())
            {
                if (typeof(ConfigurationSection).IsAssignableFrom(sectionProperty.PropertyType)
                    && sectionProperty.GetIndexParameters().Length == 0
                    && sectionProperty.PropertyType != typeof(DefaultSection))
                {
                    object propertyValue = sectionProperty.GetValue(systemWebSectionGroup, new object[0]);
                    ConfigurationSection webSection = propertyValue as ConfigurationSection;
                    if (webSection != null)
                    {
                        RecursiveBuild(webSection, systemWebNode, sectionProperty.Name);
                    }
                }
            }
        }

        private ConfigurationNode RecursiveBuild(ConfigurationElement configuration, ConfigurationNode parentNode, string nodeName)
        {
            Type configurationNodeType = nakedNodeTypeFactory.NodeTypeForDataType(configuration.GetType());
            if (configurationNodeType == typeof(NakedConfigurationNode) || configurationNodeType == parentNode.GetType()) return null;
            if (configuration is ConfigurationSection && configuration.ElementInformation.Source == null) return null;
            ConfigurationNode node = Activator.CreateInstance(configurationNodeType, configuration) as ConfigurationNode;
            if (node != null)
            {
                node.Name = nodeName;
                commandService.ForEach(parentNode.GetType(), delegate(ConfigurationUICommand command)
                {
                    if (command is AddNakedNodeConfigurationUICommand && ((AddNakedNodeConfigurationUICommand)command).ChildType == node.GetType())
                    {
                        if (!parentNode.Nodes.Contains(node))
                        {
                            parentNode.AddNode(node);
                        }
                    }
                });

                foreach (PropertyInfo property in configuration.GetType().GetProperties())
                {
                    if (typeof(ConfigurationElement).IsAssignableFrom(property.PropertyType))
                    {
                        Type childNodeType = nakedNodeTypeFactory.NodeTypeForDataType(property.PropertyType);
                        if (property.GetIndexParameters().Length == 0)
                        {
                            ConfigurationElement childConfiguration = property.GetValue(configuration, new object[0]) as ConfigurationElement;
                            if (childConfiguration != null)
                            {
                                RecursiveBuild(childConfiguration, node, property.Name);
                            }
                        }
                    }
                }
            }
            return node;
        }
    }

    public class SystemWebCommandRegistrar : NakedNodeCommandRegistrar
    {
        public SystemWebCommandRegistrar(IServiceProvider serviceProvider)
            : base(serviceProvider) { }

        public override void Register()
        {
            AddNakedNodeCommand("SystemWeb Settings", typeof(SystemWebSectionGroup), typeof(ConfigurationApplicationNode), NodeMultiplicity.Disallow, ConfigurationManager.OpenMachineConfiguration().GetSectionGroup("system.web"));
        }
    }
}
