using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Practices.EnterpriseLibrary.Configuration.Design;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Configuration;
using System.Collections;

namespace EnterpriseLibraryExtensions
{
    public class ExtenderStorageService : IExtenderStorageService
    {

        Dictionary<ExtendedPropertyKey, object> _extendedProperties = new Dictionary<ExtendedPropertyKey, object>(new ExtendedPropertyKeyEqualityComparer());
        Dictionary<IConfigurationUIHierarchy, List<UnresolvedLoadedProperty>> _unresolvedPropertiesByHierarchy = new Dictionary<IConfigurationUIHierarchy, List<UnresolvedLoadedProperty>>();
        Dictionary<IConfigurationUIHierarchy, EventHandler<ConfigurationNodeChangedEventArgs>> _unresolveHandlersByHierarchy = new Dictionary<IConfigurationUIHierarchy, EventHandler<ConfigurationNodeChangedEventArgs>>();
        IGlobalEvents _events;

        public ExtenderStorageService(IGlobalEvents events)
        {
            _events = events;
        }

        public TProperty GetExtendedPropertyValue<TProperty>(IExtenderProvider extender, string propertyName, ConfigurationNode extendee)
        {
            ExtendedPropertyKey propertyKey = new ExtendedPropertyKey(extender, propertyName, extendee);
            if (_extendedProperties.ContainsKey(propertyKey))
            {
                return (TProperty) _extendedProperties[propertyKey];
            }
            return default(TProperty);
        }

        public Dictionary<ConfigurationNode, TProperty> GetAllExtendedPropertyValues<TProperty>(IExtenderProvider extender, string propertyName)
        {
            Dictionary<ConfigurationNode, TProperty> result = new Dictionary<ConfigurationNode, TProperty>();
            foreach (ExtendedPropertyKey propertyKey in _extendedProperties.Keys)
            {
                if (propertyKey.Extender == extender && propertyKey.PropertyName == propertyName)
                {
                    result.Add(propertyKey.Extendee, (TProperty)_extendedProperties[propertyKey]);
                }
            }

            return result;
        }

        public void SetExtendedPropertyValue<TProperty>(IExtenderProvider extender, string propertyName, ConfigurationNode extendee, TProperty propertyValue)
        {
            ExtendedPropertyKey propertyKey = new ExtendedPropertyKey(extender, propertyName, extendee);
                
            if (null == (object)propertyValue)
            {
                if (_extendedProperties.ContainsKey(propertyKey))
                {
                    _extendedProperties.Remove(propertyKey);
                }
            }
            else
            {
                _extendedProperties[propertyKey] = propertyValue;
            }
        }

        public KeyValueConfigurationCollection Save<TProperty>(IExtenderProvider extender, string propertyName, Converter<TProperty, String> converter)
        {
            KeyValueConfigurationCollection result = new KeyValueConfigurationCollection();
            foreach (ExtendedPropertyKey key in _extendedProperties.Keys)
            {
                if(key.Extender == extender && string.Compare(key.PropertyName, propertyName) == 0)
                {
                    string value = null;
                    if (converter != null)
                    {
                        value = converter((TProperty)_extendedProperties[key]);
                    }
                    else
                    {
                        value = (string)Convert.ChangeType(_extendedProperties[key], typeof(string));
                    }
                    if (key.Extendee.Hierarchy != null)
                    {
                        result.Add(ApplicationNodeRelativePathForNode(key.Extendee), value);
                    }
                }
            }

            return result;
        }

        public void Load<TProperty>(IExtenderProvider extender, string propertyName, IConfigurationUIHierarchy hierarchy, KeyValueConfigurationCollection propertyValuesByNodePath, Converter<String, TProperty> converter)
        {
            string rootNodePath = hierarchy.RootNode.Path;
            foreach(KeyValueConfigurationElement propertyValueByNodePath in propertyValuesByNodePath)
            {
                ConfigurationNode foundConfigurationNode = hierarchy.FindNodeByPath(rootNodePath + propertyValueByNodePath.Key);
                if (foundConfigurationNode != null)
                {
                    TProperty value = default(TProperty);
                    if (converter != null)
                    {
                        value = converter(propertyValueByNodePath.Value);
                    } else
                    {
                        value = (TProperty) Convert.ChangeType(propertyValueByNodePath.Value, typeof(TProperty));
                    }
                    _extendedProperties.Add(new ExtendedPropertyKey(extender, propertyName, foundConfigurationNode), value);
                }
                else
                {
                    if (!_unresolvedPropertiesByHierarchy.ContainsKey(hierarchy))
                    {
                        _unresolvedPropertiesByHierarchy.Add(hierarchy, new List<UnresolvedLoadedProperty>());
                    }
                    _unresolvedPropertiesByHierarchy[hierarchy].Add( new UnresolvedLoadedProperty(propertyValueByNodePath.Key, propertyValueByNodePath.Value, propertyName, extender));
                }
            }

            if (_unresolvedPropertiesByHierarchy.ContainsKey(hierarchy) && _unresolvedPropertiesByHierarchy[hierarchy].Count > 0)
            {
                _unresolveHandlersByHierarchy[hierarchy] = new EventHandler<ConfigurationNodeChangedEventArgs>(delegate(object sender, ConfigurationNodeChangedEventArgs e) 
                { 
                    if (e.Node.Hierarchy == hierarchy)
                        TryUnresolveExtendedPropertyOnNodeAdded<TProperty>(extender, e.Node, propertyName, hierarchy, converter); 
                });
                
                _events.NodeAdded += _unresolveHandlersByHierarchy[hierarchy];
            }
        }

        void TryUnresolveExtendedPropertyOnNodeAdded<TProperty>(IExtenderProvider extender, ConfigurationNode extendee, string propertyName, IConfigurationUIHierarchy hierarchy, Converter<String, TProperty> converter)
        {
            UnresolvedLoadedProperty foundProperty = _unresolvedPropertiesByHierarchy[hierarchy].Find(delegate(UnresolvedLoadedProperty unresolvedProperty)
            {
                return extender == unresolvedProperty.Extender
                && propertyName == unresolvedProperty.PropertyName
                && (string.Compare(unresolvedProperty.ConfigurationNodePath, ApplicationNodeRelativePathForNode(extendee)) == 0);
            });

            if (foundProperty != null)
            {
                _unresolvedPropertiesByHierarchy[hierarchy].Remove(foundProperty);
                TProperty value = default(TProperty);

                if (converter != null)
                {
                    value = converter(foundProperty.PropertyValue);
                }
                else
                {
                    value = (TProperty)Convert.ChangeType(foundProperty.PropertyValue, typeof(TProperty));
                }

                _extendedProperties.Add(new ExtendedPropertyKey(extender, propertyName, extendee), value);
                if (_unresolvedPropertiesByHierarchy[hierarchy].Count == 0)
                {
                    _events.NodeAdded -= _unresolveHandlersByHierarchy[hierarchy];
                }
            }
        }

        #region class UnresolvedLoadedProperty
        private class UnresolvedLoadedProperty
        {
            string _configurationNodePath;
            string _propertyValue;
            string _propertyName;
            IExtenderProvider _extender;

            public UnresolvedLoadedProperty(string configurationNodePath, string propertyValue, string propertyName, IExtenderProvider extender)
            {
                _configurationNodePath = configurationNodePath;
                _propertyValue = propertyValue;
                _propertyName = propertyName;
                _extender = extender;
            }

            public string ConfigurationNodePath
            {
                get { return _configurationNodePath; }
            }

            public string PropertyValue
            {
                get { return _propertyValue; }
            }

            public string PropertyName
            {
                get { return _propertyName; }
            }

            public IExtenderProvider Extender
            {
                get { return _extender; }
            }
        }
        #endregion

        #region struct ExtendedPropertyKey
        private class ExtendedPropertyKeyEqualityComparer : IEqualityComparer<ExtendedPropertyKey>
        {
            public bool Equals(ExtendedPropertyKey x, ExtendedPropertyKey y)
            {
                return object.Equals(x, y);
            }

            public int GetHashCode(ExtendedPropertyKey obj)
            {
                return obj.GetHashCode();
            }
        }

        private struct ExtendedPropertyKey
        {
            IExtenderProvider _extender;
            string _propertyName;
            ConfigurationNode _extendee;

            public ExtendedPropertyKey(IExtenderProvider extender, string propertyName, ConfigurationNode extendee)
            {
                _extender = extender;
                _propertyName = propertyName;
                _extendee = extendee;
            }

            public IExtenderProvider Extender
            {
                get { return _extender; }
            }

            public string PropertyName
            {
                get { return _propertyName; }
            }

            public ConfigurationNode Extendee
            {
                get { return _extendee; }
            }
        }
        #endregion

        public static void RegisterService(IServiceProvider serviceProvider)
        {
            if (serviceProvider.GetService(typeof(IExtenderStorageService)) == null)
            {
                IServiceContainer serviceContainer = serviceProvider.GetService(typeof(IServiceContainer)) as IServiceContainer;
                if (serviceContainer != null)
                {

                    serviceContainer.AddService(typeof(IExtenderStorageService), new ExtenderStorageService(serviceProvider.GetService(typeof(IGlobalEvents)) as IGlobalEvents));
                }
            }
        }

        private string ApplicationNodeRelativePathForNode(ConfigurationNode node)
        {
            string applicationNodePath = node.Hierarchy.RootNode.Path;
            return node.Path.Substring(applicationNodePath.Length);
        }
    }

    public interface IExtenderStorageService
    {
        TProperty GetExtendedPropertyValue<TProperty>(IExtenderProvider extender, string propertyName, ConfigurationNode extendee);

        Dictionary<ConfigurationNode, TProperty> GetAllExtendedPropertyValues<TProperty>(IExtenderProvider extender, string propertyName);

        void SetExtendedPropertyValue<TProperty>(IExtenderProvider extender, string propertyName, ConfigurationNode extendee, TProperty propertyValue);

        KeyValueConfigurationCollection Save<TProperty>(IExtenderProvider extender, string propertyName, Converter<TProperty, String> converter);

        void Load<TProperty>(IExtenderProvider extender, string propertyName, IConfigurationUIHierarchy hierarchy, KeyValueConfigurationCollection propertyValuesByNodePath, Converter<String, TProperty> converter);
    }
}
